@@ -31,7 +31,7 @@ extern "C" {
 #define SEN_INDEX_DELIMITED                     0x0020
 #define SEN_INDEX_ENABLE_SUFFIX_SEARCH          0x0100
 #define SEN_INDEX_DISABLE_SUFFIX_SEARCH         0x0200
-#define SEN_INDEX_WITH_STORE                    0x1000
+#define SEN_INDEX_WITH_VGRAM                    0x1000
 #define SEN_INDEX_WITH_VACUUM                   0x8000
 
 /* 16 tokenizers can be registered */
@@ -42,6 +42,7 @@ extern "C" {
 #define SEN_SYM_WITH_SIS                        0x80000000
 
 #define SEN_SNIP_NORMALIZE                      0x0001
+#define SEN_QUERY_SCAN_NORMALIZE                SEN_SNIP_NORMALIZE
 
 #define SEN_LEX_NGRAM_UNIT_SIZE                 2
 
@@ -122,7 +123,9 @@ typedef enum {
   sen_sel_unsplit,
   sen_sel_near,
   sen_sel_similar,
-  sen_sel_term_extract
+  sen_sel_term_extract,
+  sen_sel_prefix,
+  sen_sel_suffix
 } sen_sel_mode;
 
 typedef enum {
@@ -149,7 +152,7 @@ typedef struct _sen_inv sen_inv;
 typedef struct _sen_index sen_index;
 typedef struct _sen_records sen_records;
 typedef struct _sen_set_cursor sen_set_cursor;
-typedef struct _sen_set_element * sen_set_eh;
+typedef struct _sen_set_element *sen_set_eh;
 typedef struct _sen_value sen_value;
 typedef struct _sen_values sen_values;
 typedef struct _sen_select_optarg sen_select_optarg;
@@ -161,8 +164,8 @@ typedef struct _sen_query sen_query;
 typedef struct _sen_logger_info sen_logger_info;
 typedef struct _sen_snip_mapping sen_snip_mapping;
 typedef struct _sen_records_heap sen_records_heap;
-typedef struct _sen_store sen_store;
-typedef struct _sen_store_buf sen_store_buf;
+typedef struct _sen_vgram sen_vgram;
+typedef struct _sen_vgram_buf sen_vgram_buf;
 typedef unsigned sen_id;
 
 #define SEN_ID_MAX 0x3fffffff
@@ -170,11 +173,11 @@ typedef unsigned sen_id;
 typedef sen_set_eh sen_recordh;
 
 struct _sen_index {
-  int foreign_keyp;
+  int foreign_flags;
   sen_sym *keys;
   sen_sym *lexicon;
   sen_inv *inv;
-  sen_store *store;
+  sen_vgram *vgram;
 };
 
 struct _sen_records {
@@ -191,6 +194,7 @@ struct _sen_records {
   sen_recordh *sorted;
   int ignore_deleted_records;
   void *userdata;
+  sen_id subrec_id;
 };
 
 struct _sen_value {
@@ -216,7 +220,7 @@ struct _sen_select_optarg {
 
 struct _sen_group_optarg {
   sen_sort_mode mode;
-  int (*func)(sen_records *, const sen_recordh *, void **, void *);
+  int (*func)(sen_records *, const sen_recordh *, void *, void *);
   void *func_arg;
   int key_size;
 };
@@ -256,6 +260,12 @@ typedef int (*query_term_callback)(const char *, unsigned int, void *);
 
 sen_rc sen_init(void);
 sen_rc sen_fin(void);
+sen_rc sen_info(char **version,
+                char **configure_options,
+                char **config_path,
+                sen_encoding *default_encoding,
+                unsigned int *initial_n_segments,
+                unsigned int *partial_match_threshold);
 
 sen_index *sen_index_create(const char *path, int key_size, int flags,
                             int initial_n_segments, sen_encoding encoding);
@@ -293,8 +303,8 @@ sen_records *sen_records_intersect(sen_records *a, sen_records *b);
 int sen_records_difference(sen_records *a, sen_records *b);
 sen_rc sen_records_sort(sen_records *r, int limit, sen_sort_optarg *optarg);
 sen_rc sen_records_group(sen_records *r, int limit, sen_group_optarg *optarg);
-const sen_recordh * sen_records_curr_rec(sen_records *r);
-const sen_recordh * sen_records_at(sen_records *r, const void *key,
+const sen_recordh *sen_records_curr_rec(sen_records *r);
+const sen_recordh *sen_records_at(sen_records *r, const void *key,
                                    unsigned section, unsigned pos,
                                    int *score, int *n_subrecs);
 sen_rc sen_record_info(sen_records *r, const sen_recordh *rh,
@@ -306,6 +316,10 @@ sen_rc sen_record_subrec_info(sen_records *r, const sen_recordh *rh,
 sen_index *sen_index_create_with_keys(const char *path, sen_sym *keys, int flags,
                                       int initial_n_segments, sen_encoding encoding);
 sen_index *sen_index_open_with_keys(const char *path, sen_sym *keys);
+sen_index *sen_index_create_with_keys_lexicon(const char *path, sen_sym *keys,
+                                              sen_sym *lexicon, int initial_n_segments);
+sen_index *sen_index_open_with_keys_lexicon(const char *path, sen_sym *keys,
+                                            sen_sym *lexicon);
 sen_rc sen_index_update(sen_index *i, const void *key, unsigned int section,
                         sen_values *oldvalues, sen_values *newvalues);
 sen_rc sen_index_select(sen_index *i,
@@ -319,9 +333,9 @@ sen_rc sen_index_info(sen_index *i, int *key_size, int *flags,
                       unsigned *inv_seg_size, unsigned *inv_chunk_size);
 int sen_index_path(sen_index *i, char *pathbuf, int buf_size);
 /*
-sen_set * sen_index_related_terms(sen_index *index, const char *string,
-                                  const char *(*fetcher)(void *, void *),
-                                  void *fetcher_arg);
+sen_set *sen_index_related_terms(sen_index *index, const char *string,
+                                 const char *(*fetcher)(void *, void *),
+                                 void *fetcher_arg);
 */
 
 sen_query *sen_query_open(const char *str, unsigned int str_len,
@@ -331,6 +345,8 @@ unsigned int sen_query_rest(sen_query *q, const char ** const rest);
 sen_rc sen_query_close(sen_query *q);
 sen_rc sen_query_exec(sen_index *i, sen_query *q, sen_records *r, sen_sel_operator op);
 void sen_query_term(sen_query *q, query_term_callback func, void *func_arg);
+sen_rc sen_query_scan(sen_query *q, const char **strs, unsigned int *str_lens,
+                      unsigned int nstrs, int flags, int *found, int *score);
 
 sen_rc sen_index_del(sen_index *i, const void *key);
 
@@ -354,9 +370,9 @@ sen_set *sen_set_intersect(sen_set *a, sen_set *b);
 int sen_set_difference(sen_set *a, sen_set *b);
 sen_set_eh *sen_set_sort(sen_set *set, int limit, sen_set_sort_optarg *optarg);
 
-sen_sym * sen_sym_create(const char *path, unsigned key_size,
+sen_sym *sen_sym_create(const char *path, unsigned key_size,
                          unsigned flags, sen_encoding encoding);
-sen_sym * sen_sym_open(const char *path);
+sen_sym *sen_sym_open(const char *path);
 sen_rc sen_sym_info(sen_sym *sym, int *key_size, unsigned *flags,
                     sen_encoding *encoding, unsigned *nrecords, unsigned *file_size);
 sen_rc sen_sym_close(sen_sym *sym);
@@ -374,8 +390,8 @@ sen_id sen_sym_at(sen_sym *sym, const void *key);
 sen_rc sen_sym_del(sen_sym *sym, const void *key);
 unsigned int sen_sym_size(sen_sym *sym);
 int sen_sym_key(sen_sym *sym, sen_id id, void *keybuf, int buf_size);
-sen_set * sen_sym_prefix_search(sen_sym *sym, const void *key);
-sen_set * sen_sym_suffix_search(sen_sym *sym, const void *key);
+sen_set *sen_sym_prefix_search(sen_sym *sym, const void *key);
+sen_set *sen_sym_suffix_search(sen_sym *sym, const void *key);
 sen_id sen_sym_common_prefix_search(sen_sym *sym, const void *key);
 int sen_sym_pocket_get(sen_sym *sym, sen_id id);
 sen_rc sen_sym_pocket_set(sen_sym *sym, sen_id id, unsigned int value);
@@ -433,6 +449,21 @@ if (sen_logger_pass(SEN_LOG_DEFAULT_LEVEL)) {\
   sen_logger_put(SEN_LOG_DEFAULT_LEVEL, __FILE__, __LINE__, __FUNCTION__, __VA_ARGS__);\
 }
 
+/* sen_com_client */
+
+typedef struct _sen_com_sqtp sen_com_client;
+
+sen_rc sen_com_init(void);
+void sen_com_fin(void);
+sen_com_client *sen_com_client_open(const char *host, int port);
+sen_rc sen_com_client_send(sen_com_client *c, const char *str, unsigned int str_len,
+                           int morep);
+sen_rc sen_com_client_recv(sen_com_client *c, char **str, unsigned int *str_len,
+                           int *morep, unsigned int *status, unsigned int *info);
+
+sen_rc sen_com_client_close(sen_com_client *c);
+int sen_com_client_fd(sen_com_client *c);
+
 #ifdef __cplusplus
 }
 #endif
